#include <iostream.h>
#include <strstream.h>
#include <math.h>
#include <time.h>
#include "ap.h"


// Program for testing the apfloat class, actually calculates pi

void dump (void)
{
    int i;

    cerr << "Ramsize = " << Ramsize << endl;
    cerr << "CacheL1size = " << CacheL1size << endl;
    cerr << "CacheL2size = " << CacheL2size << endl;
    cerr << "Cacheburst = " << Cacheburst << endl;
    cerr << "Maxblocksize = " << Maxblocksize << endl;
    cerr << "Memorytreshold = " << Memorytreshold << endl;
    cerr << "Blocksize = " << Blocksize << endl;
    cerr << "Cachetreshold = " << Cachetreshold << endl;
    cerr << "Cacheburstblocksize = " << Cacheburstblocksize << endl;
    cerr << "Cacheblocksize = " << Cacheblocksize << endl;
    cerr << "Base = " << Base << endl;
    cerr << "Basedigit = " << Basedigit << endl;
    cerr << "Basedigits = " << Basedigits << endl;
    cerr << "NBasefactors = " << NBasefactors << endl;
    cerr << "Basefactors = ";

    for (i = 0; i < NBasefactors; i++)
        cerr << Basefactors[i] << (i < NBasefactors - 1 ? ", " : "");

    cerr << endl;
}

long rnd2base (long x)
{
    long r = 1;

    if (!x) return 0;

    x /= Basedigits;

    while (x >>= 1) r <<= 1;

    return r * Basedigits;
}

double baseprec4 (int k, int r = 4)
{
    return (pow (4.0, (double) k) * sqrt(r) * M_PI - log (16.0 * sqrt (r)) - k * log (4.0)) / log ((double) Basedigit);
}

double baseprec2 (int k, int r = 4)
{
    return (pow (2.0, (double) k) * sqrt(r) * M_PI - log (16.0 * sqrt (r)) - k * log (2.0)) / log ((double) Basedigit);
}

apfloat pi4 (size_t destprec)
{
    int maxk = 0;
    size_t prec, dprec;
    time_t b;
    unsigned k;

    while (baseprec4 (maxk) < destprec) maxk++;

    cerr << "Using the Borweins' quartic iteration" << endl;
    cerr << "Total " << maxk << " iterations" << endl;

    cerr << "Initial values";

    b = time (0);

    apfloat y = sqrt (apfloat (2, destprec)) - 1;
    apfloat a = 2 - 4 * y;
    apfloat half = apfloat (1, destprec) / 2;
    apfloat z;

    cerr << " took " << difftime (time (0), b) << " seconds" << endl;

    for (k = 0; k < maxk; k++)
    {
        cerr << "Iteration " << k + 1;

        b = time (0);

        if (k < 2)
            z = pow (y, 4);
        else
        {
            z = y;
            prec = destprec + 4 * z.exp ();
            z.prec (prec > 1 ? prec : 1);
            z = pow (z, 4);
        }

        y = 1 - z;

        if (k < 5)
            y = invroot (y, 4);
        else
        {
            prec = rnd2base (-2 * z.exp ());
            z.prec (-z.exp ());
            z = 1 + z / 4;
            z.prec (prec);
            y = invroot (y, 4, y.prec (), z, prec);
        }

        if (k < 5)
            y = (y - 1) / (y + 1);
        else
        {
            apfloat t;

            z = y - 1;
            prec = rnd2base (-2 * z.exp ());
            dprec = z.prec ();
            t = z;
            t.prec (-z.exp ());
            t = half - t / 4;
            t.prec (prec);
            y = z * invroot (y + 1, 1, dprec, t, prec);
        }

        z = pow (y + 1, 2);
        a *= z * z;

        a -= (1UL << (2 * k + 3)) * y * (z - y);

        cerr << " took " << difftime (time (0), b) << " seconds" << endl;
    }

    cerr << "Final inverse";

    b = time (0);

    a = 1 / a;

    cerr << " took " << difftime (time (0), b) << " seconds" << endl;

    return a;
}

apfloat pi2 (size_t destprec)
{
    int maxk = 0;
    time_t m;
    unsigned k;

    while (baseprec2 (maxk) < destprec) maxk++;

    cerr << "Using the Gauss-Legendre iteration" << endl;
    cerr << "Total " << maxk << " iterations" << endl;

    cerr << "Initial values";

    m = time (0);

    apfloat b = invroot (apfloat (2, destprec), 2);
    apfloat a = apfloat (1, destprec), y;
    apfloat t = a / 4, g;

    cerr << " took " << difftime (time (0), m) << " seconds" << endl;

    for (k = 0; k < maxk; k++)
    {
        cerr << "Iteration " << k + 1;

        m = time (0);

        y = a;
        a = (a + b) / 2;

        if (k < 6)
        {
            apfloat z = b * y;
            b = 0;              // Save space
            g = invroot (z, 2);
            b = z * g;
            g.prec (rnd2base ((size_t) (200.0 / log ((double) Basedigit))));
        }
        else
        {
            apfloat z = b * y;
            b = 0;              // Save space
            g = invroot (z, 2, destprec, g);
            b = z * g;
            g.prec (rnd2base ((size_t) (200.0 / log ((double) Basedigit)) << k - 5));
        }

        if (k)
            t -= (1UL << k) * pow (y - a, 2);
        else
            t -= pow (y - a, 2);

        cerr << " took " << difftime (time (0), m) << " seconds" << endl;
    }

    cerr << "Final value";

    m = time (0);

    a = pow (a + b, 2) / (4 * t);

    cerr << " took " << difftime (time (0), m) << " seconds" << endl;

    return a;
}

int main (int argc, char *argv[])
{
    int m = 0;
    size_t prec;
    rawtype base = 10;
    time_t b;

    if (argc < 2)
    {
        cerr << "USAGE: aptest digits [method] [base]" << endl;
        cerr << "    method: 0 = Gauss-Legendre" << endl;
        cerr << "            1 = Borweins' quartic" << endl;
        cerr << "    base must be 2...36" << endl;
        return 2;
    }

    if (!((istrstream (argv[1])) >> prec))
    {
        cerr << "Invalid argument digits: " << argv[1] << endl;
        return 1;
    }

    if (argc > 2 && !((istrstream (argv[2])) >> m))
    {
        cerr << "Method must be 0 or 1" << endl;
        return 1;
    }

    if (argc > 3 && !((istrstream (argv[3])) >> base))
    {
        cerr << "Invalid argument base: " << argv[3] << endl;
        return 1;
    }

    apbase (base);
    dump ();

    cerr << "Calculating pi to " << prec << " base-" << Basedigit <<" digits" << endl;

    b = time (0);

    if (m)
        cout << pretty << pi4 (prec) << endl;
    else
        cout << pretty << pi2 (prec) << endl;

    cerr << "Total time " << difftime (time (0), b) << " seconds" << endl;

    return 0;
}
